struct DepthFirstSearch
    d_max::Int
    stack::Vector{TileIndex} # Memory used for reachability calcs
end

mutable struct DepthFirstSearchData
    solved::Bool
    max_depth::Int # max depth reached during search
    n_pushes_evaled::Int # number of pushes evaluated during search
    solve_time::Float64 # solve time in seconds
    pushes::Vector{Push} # of length d_max
    sol_depth::Int # number of pushes in the solution
end

function clear!(data::DepthFirstSearchData)
    data.solved = false
    data.max_depth = 0
    data.n_pushes_evaled = 0
    data.solve_time = 0.0
    data.sol_depth = 0
    return data
end

function solve_(dfs::DepthFirstSearch, data::DepthFirstSearchData, game::Game, s::State, reach::ReachableTiles, d::Int)

    # println(d, ":\n", board_to_text(game, s), "\n")

    data.max_depth = max(data.max_depth, d)

    if is_solved(game, s)
        # Solved!
        data.solved = true
        data.sol_depth = d
        return true
    end

    if d ≥ dfs.d_max
        # We exceeded our max depth
        return false
    end

    # Calc reachability for this state
    calc_reachable_tiles!(reach, game, s.board, s.player, dfs.stack)

    for push in get_pushes(game, s, reach)
        data.n_pushes_evaled += 1
        move!(s, game, push)
        if solve_(dfs, data, game, s, reach, d+1)
            # Store the solution
            data.pushes[d+1] = push
            return true
        end
        unmove!(s, game, push)
    end

    # Failed to find a solution
    return false
end

function solve(dfs::DepthFirstSearch, game::Game, s::State, reach::ReachableTiles)
    t_start = time()
    data = DepthFirstSearchData(false, 0, 0, 0.0, Array{Push}(undef, dfs.d_max), 0)
    clear!(reach, game.board)
    solve_(dfs, data, game, s, reach, 0)
    data.solve_time = time() - t_start
    return data
end